@@ -1,8 +1,6 @@
 from __future__ import absolute_import
 
 from loguru import logger
-from colorama import init as color_init
-from colorama import Fore, Style
 
 import multiprocessing.queues as mpq
 from multiprocessing import Process
@@ -55,12 +53,10 @@ class SharedCounter(object):
     http://eli.thegreenplace.net/2012/01/04/shared-counter-with-pythons-multiprocessing/
     """
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def __init__(self, n=0):
         self.count = multiprocessing.Value("i", n)
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def increment(self, n=1):
         """ Increment the counter by n (default = 1) """
@@ -68,7 +64,6 @@ class SharedCounter(object):
             self.count.value += n
 
     @property
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def value(self):
         """ Return the value of the counter """
@@ -99,7 +94,6 @@ class py3_mp_queue(mpq.Queue):
     # NOTE This is the new implementation based on:
     #    https://github.com/vterron/lemon/blob/master/util/queue.py
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def __init__(self, *args, **kwargs):
         # Use ctx argument if using Python3.4+
@@ -110,7 +104,6 @@ class py3_mp_queue(mpq.Queue):
         if not sys.platform == "darwin":
             self.size = SharedCounter(0)
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def put(self, *args, **kwargs):
         # Infinite recursion possible if we don't use super() here
@@ -123,7 +116,6 @@ class py3_mp_queue(mpq.Queue):
         if not sys.platform == "darwin":
             self.size.increment(1)
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def get(self, *args, **kwargs):
         if not sys.platform == "darwin":
@@ -135,7 +127,6 @@ class py3_mp_queue(mpq.Queue):
         except Empty:
             return {}
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def qsize(self):
         """ Reliable implementation of multiprocessing.Queue.qsize() """
@@ -144,7 +135,6 @@ class py3_mp_queue(mpq.Queue):
         else:
             raise NotImplementedError
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def empty(self):
         """ Reliable implementation of multiprocessing.Queue.empty() """
@@ -153,26 +143,26 @@ class py3_mp_queue(mpq.Queue):
         else:
             raise NotImplementedError
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def clear(self):
         """ Remove all elements from the Queue. """
         while not self.empty():
-            self.get()
+            self.get(block=False, timeout=0.05)
 
 
-class Worker(object):
+class Worker(multiprocessing.Process):
     """multiprocessing worker"""
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    # This is on the Worker() class
     @logger.catch(reraise=True)
-    def __init__(self, w_id, todo_q, done_q, default_sleep=0.00001):
+    def __init__(self, w_id, todo_q, done_q, default_sleep=0.00001, log_level=0):
+        super().__init__()
         assert isinstance(default_sleep, int) or isinstance(default_sleep, float)
-        color_init()
         self.w_id = w_id
         self.cycle_sleep = default_sleep  # How long the worker should sleep
         self.task = None
         self.done_q = done_q
+        self.log_level = log_level
 
         try:
             self.message_loop(todo_q, done_q)  # do work
@@ -195,7 +185,7 @@ class Worker(object):
             # time.sleep(self.cycle_sleep)
             self.task = None
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    # This is on the Worker() class
     @logger.catch(reraise=True)
     def done_q_send(self, msg_dict):
         """Send message dicts through done_q, and throw explicit errors for
@@ -212,90 +202,30 @@ class Worker(object):
             hash_func = md5()
             hash_func.update(str(msg_dict))
             dict_hash = str(hash_func.hexdigest())[-7:]  # Last 7 digits of hash
-            linesep = os.linesep
-            sys.stderr.write(
-                "{0} {1}done_q_send({2}) Can't pickle this dict:{3} '''{7}{4}   {5}{7}{6}''' {7}".format(
-                    datetime.now(),
-                    Style.BRIGHT,
-                    dict_hash,
-                    Style.RESET_ALL,
-                    Fore.MAGENTA,
-                    str(msg_dict),
-                    Style.RESET_ALL,
-                    linesep,
-                )
-            )
-            sys.stderr.write(
-                "{0}          {1}Pickling problems often come from open or hung TCP sockets{2}{3}".format(
-                    datetime.now(),
-                    Style.BRIGHT,
-                    Style.RESET_ALL,
-                    linesep,
-                )
-            )
+            logger.debug("done_q_send({0}) Can't pickle this dict:{1}".format(dict_hash, str(msg_dict),))
+            logger.debug("    Sometimes pickling problems come from open / hung TCP sockets")
 
             ## Verbose list of the offending key(s) / object attrs
             ## Send all output to stderr...
-            err_frag1 = (
-                Style.BRIGHT
-                + "    done_q_send({0}) Offending dict keys:".format(str(dict_hash))
-                + Style.RESET_ALL
-            )
-            err_frag2 = Fore.YELLOW + " {0}".format(no_pickle_keys) + Style.RESET_ALL
-            err_frag3 = "{0}".format(linesep)
-            sys.stderr.write(err_frag1 + err_frag2 + err_frag3)
+            err_frag1 = ("    done_q_send({0}) Offending dict keys:".format(str(dict_hash)))
+            err_frag2 = " {0}".format(no_pickle_keys)
+            logger.debug(err_frag1 + err_frag2)
             for key in sorted(no_pickle_keys):
-                sys.stderr.write(
-                    "      msg_dict['{0}']: {1}'{2}'{3}{4}".format(
-                        key,
-                        Fore.MAGENTA,
-                        repr(msg_dict.get(key)),
-                        Style.RESET_ALL,
-                        linesep,
-                    )
-                )
+                logger.debug("      msg_dict['{0}']: '{1}'".format(key, repr(msg_dict.get(key))))
                 if isinstance(msg_dict.get(key), object):
                     thisobj = msg_dict.get(key)
                     no_pickle_attrs = self.invalid_obj_pickle_attrs(thisobj)
-                    err_frag1 = (
-                        Style.BRIGHT
-                        + "      done_q_send({0}) Offending attrs:".format(dict_hash)
-                        + Style.RESET_ALL
-                    )
-                    err_frag2 = (
-                        Fore.YELLOW + " {0}".format(no_pickle_attrs) + Style.RESET_ALL
-                    )
-                    err_frag3 = "{0}".format(linesep)
-                    sys.stderr.write(err_frag1 + err_frag2 + err_frag3)
+                    err_frag1 = "      done_q_send({0}) Offending attrs:{1}".format(dict_hash, os.linesep)
+                    err_frag2 = " {0}".format(no_pickle_attrs)
+                    logger.debug(err_frag1 + err_frag2)
                     for attr in no_pickle_attrs:
-                        sys.stderr.write(
-                            "        msg_dict['{0}'].{1}: {2}'{3}'{4}{5}".format(
-                                key,
-                                attr,
-                                Fore.RED,
-                                repr(getattr(thisobj, attr)),
-                                Style.RESET_ALL,
-                                linesep,
-                            )
-                        )
+                        logger.debug("        msg_dict['{0}'].{1}: '{2}'".format(key, attr, repr(getattr(thisobj, attr))))
 
-            sys.stderr.write(
-                "    {0}done_q_send({1}) keys (no problems):{2}{3}".format(
-                    Style.BRIGHT, dict_hash, Style.RESET_ALL, linesep
-                )
-            )
+            logger.debug("    done_q_send({0}) keys (no problems):".format(dict_hash))
             for key in sorted(set(msg_dict.keys()).difference(no_pickle_keys)):
-                sys.stderr.write(
-                    "      msg_dict['{0}']: {1}{2}{3}{4}".format(
-                        key,
-                        Fore.GREEN,
-                        repr(msg_dict.get(key)),
-                        Style.RESET_ALL,
-                        linesep,
-                    )
-                )
+                logger.debug("      msg_dict['{0}']: {1}".format(key, repr(msg_dict.get(key))))
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    # This is on the Worker() class
     @logger.catch(reraise=True)
     def invalid_dict_pickle_keys(self, msg_dict) -> list:
         """
@@ -320,7 +250,7 @@ class Worker(object):
 
         return no_pickle_keys
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    # This is on the Worker() class
     @logger.catch(reraise=True)
     def invalid_obj_pickle_attrs(self, thisobj):
         no_pickle_attrs = []
@@ -335,14 +265,16 @@ class Worker(object):
                 no_pickle_attrs.append(attr)  # This attr is unpicklable
         return no_pickle_attrs
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    # This is on the Worker() class
     @logger.catch(reraise=True)
     def message_loop(self, todo_q, done_q):
         """Loop through messages and execute tasks"""
         t_msg = {}
         while t_msg.get("state", "") != "__DIE__":
             try:
-                t_msg = todo_q.get(True, self.cycle_sleep)  # Poll blocking
+                t_msg = todo_q.get_nowait()
+                if t_msg != {} and self.log_level >= 3:
+                    logger.debug("Got {0} from todo_q".format(t_msg))
                 self.task = t_msg.get("task", "")  # __DIE__ has no task
                 if self.task != "":
 
@@ -361,7 +293,7 @@ class Worker(object):
                     self.task.task_stop = time.time()  # Seconds since epoch
 
                     self.done_q_send(
-                        {"w_id": self.w_id, "task": self.task, "state": "__FINISHED__"}
+                        {"w_id": self.w_id, "task": self.task, "state": "__TASK_FINISHED__"}
                     )  # Ack work finished
 
                     self.task = None
@@ -369,10 +301,13 @@ class Worker(object):
                 pass
             except Full:
                 time.sleep(0.1)
+
             ## Disable extraneous error handling...
-            except Exception:
+            except Exception as eee:
                 if self.task is not None:
                     self.task.task_stop = time.time()  # Seconds since epoch
+                else:
+                    logger.error(str(eee))
                 # Handle all other errors here...
                 tb_str = "".join(tb.format_exception(*(sys.exc_info())))
                 self.done_q_send(
@@ -389,37 +324,34 @@ class Worker(object):
 
 class TaskMgrStats(object):
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def __init__(self, worker_count, log_interval=5, hot_loop=False):
-
-        assert isinstance(worker_count, int)
-        assert isinstance(log_interval, int)
         assert isinstance(hot_loop, bool)
 
-        self.worker_count = worker_count
+        self.worker_count = int(worker_count)
         self.log_interval = float(log_interval)
         self.hot_loop = hot_loop
 
-        self.stats_start = time.time()
+        self.all_tasks_started = time.time()
+        self.stats_diff_start = time.time()
         self.exec_times = []
         self.queue_times = []
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+        self.all_exec_times = []
+        self.all_queue_times = []
+
     @logger.catch(reraise=True)
     def reset_stats(self):
-        self.stats_start = time.time()
+        self.stats_diff_start = time.time()
         self.exec_times = []
         self.queue_times = []
 
     @property
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def time_delta(self):
-        return time.time() - self.stats_start
+        return time.time() - self.stats_diff_start
 
     @property
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def is_log_time(self):
         """
@@ -429,17 +361,17 @@ class TaskMgrStats(object):
             return True
         return False
 
-    @property
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
-    def log_stats_message(self):
+    def log_stats_message(self, final=False):
         """
-        Build log message strings and reset the stats
+        Build log message strings and reset the stats.  Final means all tasks are done and we need a final stats report for ALL runs.
         """
+        for ii in self.exec_times:
+            self.all_exec_times.append(ii)
+        for ii in self.queue_times:
+            self.all_queue_times.append(ii)
+
         time_delta = deepcopy(self.time_delta)
-        total_work_time = self.worker_count * time_delta
-        time_worked = sum(self.exec_times)
-        pct_busy = time_worked / total_work_time * 100.0
 
         min_task_time = min(self.exec_times)
         avg_task_time = sum(self.exec_times) / len(self.exec_times)
@@ -449,12 +381,22 @@ class TaskMgrStats(object):
         avg_queue_time = sum(self.queue_times) / len(self.queue_times)
         max_queue_time = max(self.queue_times)
 
-        time_delta = self.time_delta
-        total_tasks = len(self.exec_times)
+        if final is False:
+            time_delta = self.time_delta
+            total_tasks = len(self.exec_times)
+            time_worked = sum(self.exec_times)
+        else:
+            time_delta = time.time() - self.all_tasks_started
+            total_tasks = len(self.all_exec_times)
+            time_worked = sum(self.all_exec_times)
+
+        total_work_time = self.worker_count * time_delta
         avg_task_rate = total_tasks / time_delta
+        pct_busy = time_worked / total_work_time * 100.0
 
         # Reset stats begin time...
-        self.reset_stats()
+        if final is False:
+            self.reset_stats()
 
         stats_digits = 5
         task_msg = """Ran {0} tasks, {1} tasks/s; {2} workers {3}% busy""".format(
@@ -479,7 +421,6 @@ class TaskMgr(object):
     multiprocessing.Queue"""
 
     # http://www.jeffknupp.com/blog/2014/02/11/a-celerylike-python-task-queue-in-55-lines-of-code/
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def __init__(
         self,
@@ -520,8 +461,6 @@ class TaskMgr(object):
         self.hot_loop = hot_loop
         self.retval = set({})
 
-        color_init()
-
         self.validate_attribute_types()
 
         if hot_loop is True:
@@ -529,7 +468,6 @@ class TaskMgr(object):
             self.controller = queue
             self.supervise()  # hot_loops automatically supervise()
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def validate_attribute_types(self):
 
@@ -547,7 +485,6 @@ class TaskMgr(object):
 
         return True
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def configure_logging(self):
         """
@@ -567,12 +504,12 @@ class TaskMgr(object):
                 self.log_level = 0
             logger.enable(PACKAGE_NAME)
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def supervise(self):
         """
         If not in a hot_loop, call supervise() to start the tasks.
         """
+        state = "__WAITING__"
         self.retval = set({})
         stats = TaskMgrStats(
             worker_count=self.worker_count,
@@ -603,18 +540,17 @@ class TaskMgr(object):
 
         finished = False
         while not finished:
-
             try:
                 if hot_loop is True:
                     # Calculate the adaptive loop delay
-                    delay = self.calc_wait_time(stats.exec_times)
-                    self.queue_tasks_from_controller(delay=delay)  # queue tasks
-                    time.sleep(delay)
+                    loop_delay = self.calc_wait_time(stats.exec_times)
+                    self.queue_tasks_from_controller(loop_delay=loop_delay)  # queue tasks
+                    time.sleep(loop_delay)
 
-                r_msg = self.done_q.get_nowait()  # __ACK__ or __FINISHED__
+                r_msg = self.done_q.get_nowait()  # __ACK__ or __TASK_FINISHED__
                 task = r_msg.get("task")
                 w_id = r_msg.get("w_id")
-                state = r_msg.get("state", "")
+                state = r_msg.get("state", "__WAITING__")
 
                 # Verify that the user called `super().__init__()` in their
                 #      BaseTask() subclass __init__()
@@ -628,6 +564,7 @@ class TaskMgr(object):
                     #    in their BaseTask() subclass __init__()
                     raise NameError("You did NOT call `super().__init__()` in your BaseTask() sub-class.")
 
+
                 if state == "__ACK__":
                     self.worker_assignments[w_id] = task
                     self.work_todo.remove(task)
@@ -636,7 +573,29 @@ class TaskMgr(object):
                     if self.log_level >= 3:
                         logger.debug("w_id={0} received task={1}".format(w_id, task))
 
-                elif state == "__FINISHED__":
+                elif state == "":
+                    num_tasks = len(self.work_todo)
+                    if len(self.work_todo) > 0 or self.num_tasks_in_progress > 0:
+                        state = "__WAITING__"
+                    else:
+                        state = "__EMPTY__"
+
+                elif state == "__WAITING__":
+                    # Wating for more task updates...
+                    num_tasks = len(self.work_todo)
+                    if num_tasks > 0:
+                        continue
+
+                    elif self.num_tasks_in_progress > 0:
+                        continue
+
+                    else:
+                        # It's unclear why we are here...
+                        raise NotImplementedError
+
+                elif state == "__TASK_FINISHED__":
+                    if self.log_level >= 2:
+                        logger.info("state: __TASK_FINISHED__")
                     now = time.time()
                     task_exec_time = task.task_stop - task.task_start
                     task_queue_time = now - task.queue_time - task_exec_time
@@ -645,12 +604,11 @@ class TaskMgr(object):
 
                     if self.log_level >= 2:
                         logger.info(
-                            "TaskMgr.work_todo: {0} tasks left".format(
-                                len(self.work_todo)
+                            "TaskMgr.work_todo: {0} tasks unassigned and {1} tasks in-progress".format(
+                                len(self.work_todo), self.num_tasks_in_progress
                             )
                         )
                     if self.log_level >= 3:
-                        logger.debug("TaskMgr.work_todo: {0}".format(self.work_todo))
                         logger.debug("r_msg: {0}".format(r_msg))
 
                     if hot_loop is False:
@@ -664,7 +622,8 @@ class TaskMgr(object):
                         self.worker_assignments.pop(w_id)  # Delete the key
 
                 elif state == "__ERROR__":
-
+                    if self.log_level >= 1:
+                        logger.info("state: __ERROR__")
                     now = time.time()
                     task_exec_time = task.task_stop - task.task_start
                     task_queue_time = now - task.queue_time - task_exec_time
@@ -697,24 +656,41 @@ class TaskMgr(object):
                                 pass
                             self.retval.add(task)  # Add result to retval
 
+
                     self.respawn_dead_workers()
 
+
+                elif state == "__DIE__":
+                    # NOTE Doing NOTHING here... why??
+                    if self.log_level >= 1:
+                        logger.info("state: __DIE__")
+
+
+                else:
+                    if len(self.work_todo) > 0:
+                        logger.error("Unhandled state: '{1}' with {0} tasks todo".format(num_tasks, state))
+                        raise ValueError("Unknown state: '{0}'".format(state))
+                    else:
+                        logger.error("Unhandled state: '{1}' with {0} tasks todo".format(num_tasks, state))
+
             except Empty:
                 state = "__EMPTY__"
+                if self.log_level >= 1:
+                    logger.info("state: '__EMPTY__'")
 
             except Exception as ee:
                 tb_str = "".join(tb.format_exception(*(sys.exc_info())))
-                print("ERROR:")
-                print(ee, tb_str)
+                logger.error("ERROR: " + str(ee) + str(tb_str))
 
-            # PIZZA
             if stats.is_log_time is True:
                 if self.log_level >= 0:
-                    logger.info(stats.log_stats_message)
+                    logger.info(stats.log_stats_message(final=False))
 
             # Adaptive loop delay unless on Mac OSX... OSX delay is constant...
-            delay = self.calc_wait_time(stats.exec_times)
-            time.sleep(delay)
+
+            # At most, sleep 50 milliseconds...
+            loop_delay = self.calc_wait_time(stats.exec_times)
+            time.sleep(loop_delay)
 
             self.respawn_dead_workers()
             finished = self.are_tasks_finished()
@@ -723,77 +699,81 @@ class TaskMgr(object):
         if hot_loop is False:
             self.stop_workers()
             for w_id, p in self.workers.items():
+                if self.log_level >= 1:
+                    logger.info("Stopping {0} with mp.join()".format(str(p)))
                 p.join()
 
-            # PIZZA
             ## Log a final stats summary...
             if self.log_level > 0:
-                logger.info(stats.log_stats_message)
+                logger.info(stats.log_stats_message(final=True))
             return self.retval
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def calc_wait_time(self, exec_times):
         num_samples = float(len(exec_times))
 
         # Work around Mac OSX problematic queue.qsize() implementation...
         if sys.platform == "darwin":
-            wait_time = 0.00001  # 10us delay to avoid worker / done_q race
+            wait_time = 0.00001  # 10us loop_delay to avoid worker / done_q race
 
         elif num_samples > 0.0:
             # NOTE:  OSX doesn't implement queue.qsize(), I worked around the
             # problem
             queue_size = max(self.done_q.qsize(), 1.0) + max(self.todo_q.qsize(), 1.0)
             min_task_time = min(exec_times)
-            wait_time = min_task_time / queue_size
+            wait_time = float(min_task_time) / float(queue_size)
 
         else:
-            wait_time = 0.00001  # 10us delay to avoid worker / done_q race
+            wait_time = 0.00001  # 10us loop_delay to avoid worker / done_q race
 
-        return wait_time
+        # Wait at MOST 1-millisecond...
+        return min(0.001, wait_time)
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
-    def queue_tasks_from_controller(self, delay=0.0):
+    def queue_tasks_from_controller(self, loop_delay=0.0):
         finished = False
         while not finished:
             try:
                 ## Hot loops will queue a list of tasks...
                 tasklist = self.controller.to_taskmgr_q.get_nowait()
                 for task in tasklist:
-                    if delay > 0.0:
-                        task.worker_loop_delay = delay
+                    if loop_delay > 0.0:
+                        task.worker_loop_delay = loop_delay
                     self.work_todo.append(task)
                     self.queue_task(task)
             except Empty:
                 # Poll until empty
                 finished = True
-            # except (Exception) as e:
-            #    tb.print_exc()
+            except Exception:
+                tb.print_exc()
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def queue_task(self, task):
         task.queue_time = time.time()  # Record the queueing time
         self.todo_q_send({"task": task})
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @property
+    @logger.catch(reraise=True)
+    def num_tasks_in_progress(self):
+        """Number of assigned tasks that are still in-progress."""
+        return len(self.worker_assignments.keys())
+
     @logger.catch(reraise=True)
     def are_tasks_finished(self):
-        if (len(self.work_todo) == 0) and (len(self.worker_assignments.keys()) == 0):
+        if (len(self.work_todo) == 0) and self.num_tasks_in_progress == 0:
             return True
+
         elif (self.hot_loop is False) and (len(self.retval)) == self.num_tasks:
             # We need this exit condition due to __ERROR__ race conditions...
             return True
+
         return False
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def stop_workers(self):
         stop = {"state": "__DIE__"}
         [self.todo_q_send(stop) for x in range(0, self.worker_count)]
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def respawn_dead_workers(self):
         """Respawn workers / tasks upon crash"""
@@ -828,12 +808,11 @@ class TaskMgr(object):
                     )
                 self.workers[w_id] = Process(
                     target=Worker,
-                    args=(w_id, self.todo_q, self.done_q, self.worker_cycle_sleep),
+                    args=(w_id, self.todo_q, self.done_q, self.worker_cycle_sleep, self.log_level),
                 )
                 self.workers[w_id].daemon = True
                 self.workers[w_id].start()
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def spawn_workers(self):
         workers = {}
@@ -847,7 +826,6 @@ class TaskMgr(object):
             workers[w_id].start()
         return workers
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def todo_q_send(self, msg_dict):
 
@@ -862,81 +840,30 @@ class TaskMgr(object):
             hash_func.update(str(msg_dict))
             dict_hash = str(hash_func.hexdigest())[-7:]  # Last 7 digits of hash
             linesep = os.linesep
-            sys.stderr.write(
-                "{0} {1}todo_q_send({2}) Can't pickle this dict:{3} '''{7}{4}   {5}{7}{6}''' {7}".format(
-                    datetime.now(),
-                    Style.BRIGHT,
-                    dict_hash,
-                    Style.RESET_ALL,
-                    Fore.MAGENTA,
-                    msg_dict,
-                    Style.RESET_ALL,
-                    linesep,
-                )
+            logger.warning(
+                "in todo_q_send({0}) Can't pickle this dict:'''{1}'''{2}".format(dict_hash, msg_dict, os.linesep)
             )
 
-            ## Verbose list of the offending key(s) / object attrs
-            ## Send all output to stderr...
-            err_frag1 = (
-                Style.BRIGHT
-                + "    todo_q_send({0}) Offending dict keys:".format(dict_hash)
-                + Style.RESET_ALL
-            )
-            err_frag2 = Fore.YELLOW + " {0}".format(no_pickle_keys) + Style.RESET_ALL
-            err_frag3 = "{0}".format(linesep)
-            sys.stderr.write(err_frag1 + err_frag2 + err_frag3)
+            ## Verbose list of the offending key(s) / object attrs...
+            err_frag1 = "    todo_q_send({0}) Offending dict keys: {1}{2}".format(dict_hash, no_pickle_keys, os.linesep)
+            err_frag2 = " {0}{1}".format(no_pickle_keys, os.linesep)
+            logger.warning(err_frag1 + err_frag2)
+
             for key in sorted(no_pickle_keys):
-                sys.stderr.write(
-                    "      msg_dict['{0}']: {1}'{2}'{3}{4}".format(
-                        key,
-                        Fore.MAGENTA,
-                        repr(msg_dict.get(key)),
-                        Style.RESET_ALL,
-                        linesep,
-                    )
-                )
+                logger.debug("      msg_dict['{0}']: '{1}'".format(key, repr(msg_dict.get(key))))
                 if isinstance(msg_dict.get(key), object):
                     thisobj = msg_dict.get(key)
                     no_pickle_attrs = self.invalid_obj_pickle_attrs(thisobj)
-                    err_frag1 = (
-                        Style.BRIGHT
-                        + "      todo_q_send({0}) Offending attrs:".format(dict_hash)
-                        + Style.RESET_ALL
-                    )
-                    err_frag2 = (
-                        Fore.YELLOW + " {0}".format(no_pickle_attrs) + Style.RESET_ALL
-                    )
-                    err_frag3 = "{0}".format(linesep)
-                    sys.stderr.write(err_frag1 + err_frag2 + err_frag3)
+                    err_frag1 = "      todo_q_send({0}) Offending attrs:".format(dict_hash)
+                    err_frag2 = " {0}".format(no_pickle_attrs)
+                    logger.warning(err_frag1 + err_frag2)
                     for attr in no_pickle_attrs:
-                        sys.stderr.write(
-                            "        msg_dict['{0}'].{1}: {2}'{3}'{4}{5}".format(
-                                key,
-                                attr,
-                                Fore.RED,
-                                repr(getattr(thisobj, attr)),
-                                Style.RESET_ALL,
-                                linesep,
-                            )
-                        )
+                        logger.debug("        msg_dict['{0}'].{1}: '{2}'".format(key, attr, repr(getattr(thisobj, attr))))
 
-            sys.stderr.write(
-                "    {0}todo_q_send({1}) keys (no problems):{2}{3}".format(
-                    Style.BRIGHT, dict_hash, Style.RESET_ALL, linesep
-                )
-            )
+            logger.debug("    {0}todo_q_send({1}) keys (no problems):{2}{3}".format(dict_hash))
             for key in sorted(set(msg_dict.keys()).difference(no_pickle_keys)):
-                sys.stderr.write(
-                    "      msg_dict['{0}']: {1}{2}{3}{4}".format(
-                        key,
-                        Fore.GREEN,
-                        repr(msg_dict.get(key)),
-                        Style.RESET_ALL,
-                        linesep,
-                    )
-                )
+                logger.debug("      msg_dict['{0}']: {1}".format(key, repr(msg_dict.get(key))))
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def invalid_dict_pickle_keys(self, msg_dict):
         """Return a list of keys that can't be pickled.  Return [] if
@@ -955,7 +882,6 @@ class TaskMgr(object):
                 no_pickle_keys.append(key)  # This key has an unpicklable value
         return no_pickle_keys
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def invalid_obj_pickle_attrs(self, thisobj):
         no_pickle_attrs = []
@@ -974,7 +900,6 @@ class TaskMgr(object):
 class ControllerQueue(object):
     """A set of queues to manage a continuous hot TaskMgr work loop"""
 
-    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
     @logger.catch(reraise=True)
     def __init__(self):
         ## to and from are with respect to the (client) controller object
